@@ -3,7 +3,10 @@ __all__ = []
 
 
 from pyiron_atomistics import Project
+
+
 from pyiron_base import JOB_CLASS_DICT
+from pyiron_base import add_module_conversion
 
 
 # Make classes available for new pyiron version
@@ -21,10 +24,19 @@ JOB_CLASS_DICT.update(
         "PacemakerJob": "pyiron_potentialfit.pacemaker.job",
         "MeamFit": "pyiron_potentialfit.meamfit.meamfit",
         "FitsnapJob": "pyiron_potentialfit.fitsnap.job",
+        "AssystStructures": "pyiron_potentialfit.assyst.structures.job",
     }
 )
 
 
+# for module in moved_potential_modules:
+add_module_conversion(
+    "pyiron_potentialfit.spgfit.structures", "pyiron_potentialfit.assyst.structures"
+)
+add_module_conversion(
+    "pyiron_potentialfit.spgfit.projectflow", "pyiron_potentialfit.assyst.projectflow"
+)
+
 from ._version import get_versions
 
 __version__ = get_versions()["version"]
