@@ -29,6 +29,7 @@ Fe_bcc  ...
 """
 
 from typing import Callable, Dict, Any, Optional
+from collections import Counter
 
 import numpy as np
 import pandas as pd
@@ -60,7 +61,7 @@ class TrainingContainer(GenericJob, HasStructure):
             {"save_neighbors": True, "num_neighbors": 12}, table_name="parameters"
         )
 
-    def include_job(self, job, iteration_step=-1):
+    def include_job(self, job, iteration_step=-1, identifier=None):
         """
         Add structure, energy and forces from job.
 
@@ -68,8 +69,9 @@ class TrainingContainer(GenericJob, HasStructure):
             job (:class:`.AtomisticGenericJob`): job to take structure from
             iteration_step (int, optional): if job has multiple steps, this
             selects which to add
+            identifier (str): name of the extracted structure in the container; default is the name of the job
         """
-        self._container.include_job(job, iteration_step)
+        self._container.include_job(job, iteration_step, identifier=identifier)
 
     def include_structure(self, structure, energy=None, name=None, **properties):
         """
@@ -293,6 +295,22 @@ class TrainingContainer(GenericJob, HasStructure):
         """
         yield from self._container.iter(*arrays, wrap_atoms=wrap_atoms)
 
+    def subtract_reference(self, reference_energies: dict[str, float]):
+        """
+        Subtract Atomic energies from structure energies.
+
+        Must be called before the job is run!
+        The original energy is kept in a new array "energy_uncorrected".
+
+        Args:
+            reference_energies (dict): a dictionary from element symbols to the energy of their isolated atoms.
+        """
+        if not self.status.initialized:
+            raise ValueError(
+                f"Must be called before job is run, not in state: '{self.status}'!"
+            )
+        self._container.subtract_reference(reference_energies)
+
 
 class TrainingPlots(StructurePlots):
     """
@@ -377,6 +395,7 @@ _HDF_KEYS = {
     "cell": "output/generic/cells",
     "positions": "output/generic/positions",
     "pbc": "input/structure/cell/pbc",
+    "spins": "output/generic/dft/final_magmoms",
 }
 
 _JOB_HDF_OVERLAY_KEYS = {
@@ -440,6 +459,7 @@ class TrainingStorage(StructureStorage):
         job,
         iteration_step=-1,
         hdf_keys=None,
+        identifier=None,
     ):
         """
         Add structure, energy, forces and pressures from an inspected or loaded job.
@@ -461,6 +481,7 @@ class TrainingStorage(StructureStorage):
         * `positions`: `output/generic/positions`
         * `species`: `input/structure/species`
         * `pbc`: `input/structure/cell/pbc`
+        * `spins`: `output/generic/dft/final_magmoms`
 
         Other keys are ignored.  All entries except `pbc` and `species` must
         lead to arrays that can be indexed by `iteration_step`.
@@ -477,49 +498,67 @@ class TrainingStorage(StructureStorage):
             iteration_step (int, optional): if job has multiple steps, this selects which to add
             hdf_keys (dict of str): customize where values are read from the
                                     job HDF5 file
+            identifier (str): name of the extracted structure in the storage; default is the name of the job
         """
 
         hdf_keys = _HDF_KEYS.copy()
-        hdf_keys.update(_JOB_HDF_OVERLAY_KEYS.get(job["NAME"], {}))
+        hdf_keys.update(_JOB_HDF_OVERLAY_KEYS.get(job.content["NAME"], {}))
 
         kwargs = {
-            "energy": job[hdf_keys["energy"]][iteration_step],
+            "energy": job.content[hdf_keys["energy"]][iteration_step],
         }
-        ff = job[hdf_keys["forces"]]
-        if ff is not None:
-            kwargs["forces"] = ff[iteration_step]
+        try:
+            kwargs["forces"] = job.content[hdf_keys["forces"]][iteration_step]
+        except KeyError:
+            pass
 
-        pp = job[hdf_keys["stress"]]
-        if pp is not None and len(pp) > 0:
-            stress = np.asarray(pp[iteration_step])
-            if stress.shape == (3, 3):
-                stress = np.array(
-                    [
-                        stress[0, 0],
-                        stress[1, 1],
-                        stress[2, 2],
-                        stress[1, 2],
-                        stress[0, 2],
-                        stress[0, 1],
-                    ]
-                )
-            kwargs["stress"] = stress
+        try:
+            pp = job.content[hdf_keys["stress"]]
+            if len(pp) > 0:
+                stress = np.asarray(pp[iteration_step])
+                if stress.shape == (3, 3):
+                    stress = np.array(
+                        [
+                            stress[0, 0],
+                            stress[1, 1],
+                            stress[2, 2],
+                            stress[1, 2],
+                            stress[0, 2],
+                            stress[0, 1],
+                        ]
+                    )
+                kwargs["stress"] = stress
+        except KeyError:
+            pass
 
-        ii = job[hdf_keys["indices"]]
-        if ii is not None:
-            indices = ii[iteration_step]
-        else:
+        try:
+            indices = job.content[hdf_keys["indices"]][iteration_step]
+        except KeyError:
             # not all jobs save indices again in the output, but all atomistic
             # jobs do it in the input
-            indices = job["input/structure/indices"]
-        species = np.asarray(job[hdf_keys["species"]])
-        cell = job[hdf_keys["cell"]][iteration_step]
-        positions = job[hdf_keys["positions"]][iteration_step]
-        pbc = job[hdf_keys["pbc"]]
+            indices = job.content["input/structure/indices"]
+        species = np.asarray(job.content[hdf_keys["species"]])
+        cell = job.content[hdf_keys["cell"]][iteration_step]
+        positions = job.content[hdf_keys["positions"]][iteration_step]
+        pbc = job.content[hdf_keys["pbc"]]
+        try:
+            kwargs["spins"] = np.array(job.content[hdf_keys["spins"]][iteration_step])
+        except (KeyError, IndexError):
+            pass
+
+        if "forces" in kwargs and not self.has_array("forces"):
+            self.add_array(
+                "forces", shape=(3,), dtype=np.float64, per="element", fill=np.nan
+            )
+        if "stress" in kwargs and not self.has_array("stress"):
+            # save stress in voigt notation
+            self.add_array(
+                "stress", shape=(6,), dtype=np.float64, per="chunk", fill=np.nan
+            )
 
         self.add_chunk(
             len(indices),
-            identifier=job.name,
+            identifier=identifier or job.name,
             symbols=species[indices],
             positions=positions,
             cell=[cell],
@@ -658,3 +697,26 @@ class TrainingStorage(StructureStorage):
         if self._plots is None:
             self._plots = TrainingPlots(self)
         return self._plots
+
+    def subtract_reference(self, reference_energies: dict[str, float]):
+        """
+        Subtract Atomic energies from structure energies.
+
+        The original energy is kept in a new array "energy_uncorrected".
+
+        Args:
+            reference_energies (dict): a dictionary from element symbols to the energy of their isolated atoms.
+        """
+        elements = self.get_elements()
+        if not set(reference_energies).issuperset(elements):
+            raise ValueError(
+                f"Must specify reference energies for all present elements: {elements}!"
+            )
+        counts = list(map(Counter, self.get_array_ragged("symbols")))
+        bias = np.array(
+            [sum(reference_energies[e] * c[e] for e in elements) for c in counts]
+        )
+        self._per_chunk_arrays["energy_uncorrected"] = self._per_chunk_arrays[
+            "energy"
+        ].copy()
+        self._per_chunk_arrays["energy"][: self.num_chunks] -= bias
